//*******************************************************
//
//	ワイドバンド AM/FMラジオ / サイレントシステム SR-1
//
//	ATmega328p
//		efuse:0xFD	BOD 2.7V
//		hfuse:0xD1	EEPROM消去保護
//		lfuse:0xE2	内蔵CR発振 8MHz
//
//	Compiler:	AVR-GCC		4.3.4
//				AVR-LIBC	1.6.7
//				binutils	2.20
//
//	2010.11.30	KAWAKAMI Yukio
//
//	New BSD license
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>


#define		F_CPU	8000000UL
#include	<util/delay.h>

#include	"radio.h"
#include	"lcd.h"
#include	"com.h"

#define		ENC_NOP		0
#define		ENC_UP		1
#define		ENC_DOWN	2

PROGMEM uchar EncData[] = {		// ロータリーエンコーダの動作シーケンス
	ENC_NOP,	ENC_UP,		ENC_DOWN,	ENC_NOP,	// 0 → 0, 1, 2, 3
	ENC_DOWN,	ENC_NOP,	ENC_NOP,	ENC_UP,		// 1 → 0, 1, 2, 3
	ENC_UP,		ENC_NOP,	ENC_NOP,	ENC_DOWN,	// 2 → 0, 1, 2, 3
	ENC_NOP,	ENC_DOWN,	ENC_UP,		ENC_NOP		// 3 → 0, 1, 2, 3
};

char	Band;
char	StepAM;
char	StepFM;
int		FreqMW;
int		FreqAM;
int		FreqFM;

volatile uchar	Timer;
volatile char	SecFlag;
volatile ushort	VoltData;
volatile ushort	VolumeData;

//*******************************************************
//	タイマー２比較一致割り込み
//	8MHz/1024/78≒100Hz
//*******************************************************
ISR(TIMER2_COMPA_vect){
	static char	sec_count = 0;

	if (++sec_count >= 100){
		sec_count = 0;
		SecFlag = 1;
	}

	if (Timer)		--Timer;		// 汎用タイマ

	// A/D変換開始 プリスケーラ 8MHz/64 = 125kHz
	ADMUX = (ADMUX & 0xF0) | VOL;
	ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADIE)|(1<<ADPS1)|(1<<ADPS2);
}


//*******************************************************
//	A/D変換の割り込みルーチン
//*******************************************************
ISR(ADC_vect) {

	if ((ADMUX & 0x0F) == VOL){
		VolumeData = (ushort)ADCL;
		VolumeData |= (((ushort)ADCH & 3) << 8);
		ADMUX = (ADMUX & 0xF0)|BATT_VOLT;
		ADCSRA = (1<<ADEN)|(1<<ADSC)|(1<<ADIE)|(1<<ADPS1)|(1<<ADPS2);
	} else {
		VoltData = (ushort)ADCL;
		VoltData |= (((ushort)ADCH & 3) << 8);
		ADCSRA = 0;
	}
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main(void){
	int		old_vol = -1000;
	char	vol = -1;
	int		*freq;
	int		freq_step;
	int		freq_min;
	int		freq_max;
	int		old_freq = -1;
	char	buff[32];

	cli();

	PORTD = (1<<ENC2)|(1<<ENC1)|(1<<SW6)|(1<<SW5)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<PLOCK)|(1<<AMP);

	PORTC = (1<<SW4)|(1<<SW3)|(1<<SW2)|(1<<SW1);
	DDRC = (1<<LED)|(1<<ANT);

	PORTB = (1<<NC1)|(1<<NC2)|(1<<CS1);
	DDRB = (1<<SCK)|(1<<MOSI)|(1<<CS1)|(1<<A0)|(1<<LRES);

	if (MCUSR & (1<<WDRF)){
		// ウォッチドッグリセット
	}
	MCUSR = 0;

	WDR;
	WDTCSR = (1<<WDCE)|(1<<WDE);	// WDCEとWDEを1にしてタイマー変更をする
	WDTCSR = (1<<WDE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);	// ウォッチドッグタイマーセット 約２秒

	ADMUX = (1<<REFS1)|(1<<REFS0);		// 内部基準電圧 1.1V
	ADCSRA = (1<<ADEN)|(1<<ADPS2)|(1<<ADPS1);	// A/D変換回路ON プリスケーラ 8MHz/64 = 125kHz

	TCCR2A = (1<<WGM21);					// CTC動作
	TCCR2B = (1<<CS22)|(1<<CS21)|(1<<CS20);	// プリスケーラ 1/1024
	OCR2A = 78 - 1;			// タイマカウント最大値  8MHz/1024/78≒100Hz
	TIMSK2 = (1<<OCIE2A);	// 比較Ａ割り込み許可

	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// TWI, TIMER0, TIMER1 電力カット
	PRR = (1<<PRTWI)|(1<<PRTIM0)|(1<<PRTIM1);
	ACSR = (1<<ACD);			// アナログコンパレータ禁止

	SPCR = (1<<SPE)|(1<<MSTR);	// SPIマスタ Mode 0
	SPSR = 0;			// AD-12864-SPIは 2.5MHz maxらしい

	sei();		// 割り込み許可

	wait_timer(50);
	
	PORTD |= (1<<PLOCK);	// 電源保持

	PORTC |= (1<<LED);

	LCD_init();
	print_P(PSTR("Wide Band AM/FM Radio"));
	print_P(PSTR("\nVersion " VERSION "\nhttp://yuki-lab.jp/"));

	// クロック補正 CPU個体差があるので要調整
	// 57600bpsはプラス誤差になるのでクロックを少し遅くする
	OSCCAL = OSCCAL - 12;
	init_com(57600);

	Band = (char)eeprom_read_byte(EEPROM_BAND);
	if (Band == 0xFF)	Band = BAND_MW;
	StepAM = (char)eeprom_read_byte(EEPROM_STEP_AM);
	if (StepAM == 0xFF)	StepAM = STEP_MIDDLE;
	StepFM = (char)eeprom_read_byte(EEPROM_STEP_FM);
	if (StepFM == 0xFF)	StepFM = STEP_MIDDLE;
	FreqMW = (uint)eeprom_read_word(EEPROM_FREQ_MW);
	if (FreqMW == 0xFFFF)	FreqMW = 1062;
	FreqAM = (uint)eeprom_read_word(EEPROM_FREQ_AM);
	if (FreqAM == 0xFFFF)	FreqAM = 6055;
	FreqFM = (uint)eeprom_read_word(EEPROM_FREQ_FM);
	if (FreqFM == 0xFFFF)	FreqFM = 8230;
	
	wait_timer(50);

	// 電源スイッチが放されるまで待つ
	wait_sw_off();

	clear_LCD();
	print_P(PSTR("Wide Band AM/FM Radio"));
	PORTD |= (1<<AMP);	// AMP ON

	read_encorder();
	
	freq = set_band(Band, &freq_step, &freq_min, &freq_max);

	for(;;){
		uchar	sw;
		int		vol_data;

		wait_timer(1);

		sw = check_sw();

		switch(sw){
		  case SW_POWER:
			power_off();	// 電源OFF
			break;

		  case SW_MODE:
			if (++Band > BAND_FM)	Band = BAND_MW;
			freq = set_band(Band, &freq_step, &freq_min, &freq_max);
			wait_sw_off();
			break;

		  case SW_STEP:
			if (Band == BAND_AM){
				if (++StepAM > STEP_HIGH)	StepAM = STEP_LOW;
				freq = set_band(Band, &freq_step, &freq_min, &freq_max);
			} else if (Band == BAND_FM){
				if (++StepFM > STEP_HIGH)	StepFM = STEP_LOW;
				freq = set_band(Band, &freq_step, &freq_min, &freq_max);
			}
			wait_sw_off();
			break;

		  default:
			break;
		}
		
		if (SecFlag){
			int v;
			v = read_batt();
			if (v <= 2.05){
				power_off();	// 電源OFF
			}
			SecFlag = 0;
			itoa(read_batt(), buff, 10);
			buff[5] = 0;
			buff[4] = 'V';
			buff[3] = buff[2];
			buff[2] = buff[1];
			buff[1] = '.';
			locate(8, 7);
			print_P(PSTR("Batt:"));
			print(buff);
			//while(check_com())	read_com();
			//write_str_P(PSTR("c\r"));
			//locate(0, 4);
			//print_response();
		}

		cli();
		vol_data = VolumeData;
		sei();
		if (abs(vol_data - old_vol) > 16){
			uchar	v;
			old_vol = vol_data;
			v = (vol_data + 16) >> 5;
			if (vol != v){
				vol = v;
				itoa(vol, buff, 10);
				write_com('v');
				write_str((uchar *)buff);
				write_com(CR);
				wait_prompt();
				locate(0, 7);
				print_P(PSTR("vol:"));
				print(buff);
				putch(' ');
			}
		}

		switch(read_encorder()){
		  case ENC_UP:
			if (*freq < freq_max){
				*freq += freq_step;
			}
			if (*freq % freq_step){
				*freq /= freq_step;
				*freq *= freq_step;
			}
			break;
		  case ENC_DOWN:
			if (*freq > freq_min){
				*freq -= freq_step;
			}
			if (*freq % freq_step){
				*freq /= freq_step;
				*freq *= freq_step;
				*freq += freq_step;
			}
			break;
		}

		if (*freq != old_freq){
			itoa(*freq, buff, 10);
			write_com('t');
			write_str((uchar *)buff);
			write_com(CR);
			wait_prompt();
			if (Band == BAND_FM){
				strcpy_P(buff, PSTR("FM"));
				print_large_num(0, 16, buff);
				sprintf(buff, "%3d.%02dMHz", *freq / 100, *freq % 100);
			} else if (Band == BAND_AM){
				strcpy_P(buff, PSTR("AM"));
				print_large_num(0, 16, buff);
				sprintf(buff, "%2d.%03dMHz", *freq / 1000, *freq % 1000);
			} else {
				strcpy_P(buff, PSTR("MW"));
				print_large_num(0, 16, buff);
				sprintf(buff, "  %4dkHz", *freq);
			}
			print_large_num(28, 16, buff);
			old_freq = *freq;
		}
	}
}


//*******************************************************
//	MW/AM/FM切り替え
//*******************************************************
int *set_band(char band, int *step, int *min, int *max){
	int *freq = NULL;
	static char	mode = 0;

	locate(0, 6);

	switch(band){
	  case BAND_MW:
		print_P(PSTR("Step:   9kHz"));
		*step = 9;
		*min = 522;
		*max = 1602;
		freq = &FreqMW;
		if (mode != 2){
			write_str_P(PSTR("b2\r"));	// AM
			wait_prompt();
			mode = 2;
		}
		break;
	  case BAND_AM:
		switch(StepAM){
		  case STEP_LOW:
			*step = 1;
			print_P(PSTR("Step:   1kHz"));
			break;
		  case STEP_MIDDLE:
			*step = 5;
			print_P(PSTR("Step:   5kHz"));
			break;
		  case STEP_HIGH:
			*step = 100;
			print_P(PSTR("Step: 100kHz"));
			break;
		}
		*min = 150;
		*max = 23000;
		freq = &FreqAM;
		if (mode != 2){
			write_str_P(PSTR("b2\r"));	// AM
			wait_prompt();
			mode = 2;
		}
		break;
	  case BAND_FM:
		switch(StepFM){
		  case STEP_LOW:
			print_P(PSTR("Step:0.01MHz"));
			*step = 1;
			break;
		  case STEP_MIDDLE:
			print_P(PSTR("Step: 0.1MHz"));
			*step = 10;
			break;
		  case STEP_HIGH:
			print_P(PSTR("Step:   1MHz"));
			*step = 100;
			break;
		}
		*min = 6400;
		*max = 10800;
		freq = &FreqFM;
		if (mode != 1){
			write_str_P(PSTR("b1\r"));	// FM
			wait_prompt();
			mode = 1;
		}
		break;
	}

	return freq;
}


//*******************************************************
//	スイッチ入力チェック
//*******************************************************
uchar check_sw(void){

	if (!(PINC & (1<<SW1)))	return SW_POWER;
	if (!(PINC & (1<<SW2)))	return SW_MODE;
	if (!(PINC & (1<<SW3)))	return SW_SW3;
	if (!(PINC & (1<<SW4)))	return SW_SW4;
	if (!(PIND & (1<<SW5)))	return SW_SW5;
	if (!(PIND & (1<<SW6)))	return SW_STEP;
	return 0;
}


//*******************************************************
//	スイッチが放されるまで待つ
//*******************************************************
void wait_sw_off(void){

	WDR;
	SLEEP;
	while(check_sw()){
		WDR;
		SLEEP;
	}
}


//*******************************************************
// ロータリーエンコーダ読み込み
//*******************************************************
uchar read_encorder(void){
	static uchar	enc = 0;
	uchar	ret;

	enc = ((enc & 3) << 2) | ((PIND>>ENC1) & 3);

	ret = pgm_read_byte(&EncData[enc]);
	if (ret == ENC_NOP)	return ret;
	if (((enc & 3) == 0)||((enc & 3) == 3)){
		return ret;
	} else {
		return ENC_NOP;
	}
}


//*******************************************************
//	SR-1 レスポンス表示
//*******************************************************
void print_response(void){
	uchar	c;

	WDR;
	Timer = 20;	// max 200ms

	while(Timer){
		if (check_com()){
			if(read_com() == LF)	break;
		}
	}
	while(Timer){
		if (check_com()){
			c = read_com();
			if (c == '>')	break;
			putch(c);
		}
	}
}


//*******************************************************
//	SR-1 プロンプト待ち
//*******************************************************
void wait_prompt(void){

	Timer = 20;		// max 100ms
	while(Timer){
		WDR;
		SLEEP;
		if (check_com()){
			if (read_com() == '>')	break;
		}
	}
}


//*******************************************************
//	電源を切る
//*******************************************************
void power_off(void){

	PORTD &= ~(1<<AMP);		// AMP OFF
	LCD_off();

	write_str_P(PSTR("b0\r"));
	wait_prompt();

	if (eeprom_read_byte(EEPROM_BAND) != Band){
		eeprom_write_byte(EEPROM_BAND, Band);
	}
	if (eeprom_read_byte(EEPROM_STEP_AM) != StepAM){
		eeprom_write_byte(EEPROM_STEP_AM, StepAM);
	}
	if (eeprom_read_byte(EEPROM_STEP_FM) != StepFM){
		eeprom_write_byte(EEPROM_STEP_FM, StepFM);
	}
	if (eeprom_read_word(EEPROM_FREQ_MW) != FreqMW){
		eeprom_write_word(EEPROM_FREQ_MW, FreqMW);
	}
	if (eeprom_read_word(EEPROM_FREQ_AM) != FreqAM){
		eeprom_write_word(EEPROM_FREQ_AM, FreqAM);
	}
	if (eeprom_read_word(EEPROM_FREQ_FM) != FreqFM){
		eeprom_write_word(EEPROM_FREQ_FM, FreqFM);
	}
	
	PORTD &= ~(1<<PLOCK);	// 電源OFF

	for(;;){
		WDR;
		SLEEP;
	}
}


//*******************************************************
//	音量設定読み込み
//*******************************************************
uchar read_vol(void){
	ushort	data;

	cli();
	data = VolumeData;
	sei();

	data += 2;
	data >>= 2;
	return (uchar)data;
}


//*******************************************************
//	バッテリー電圧読み込み
//*******************************************************
int read_batt(void){
	long	l;

	cli();
	l = (long)VoltData;
	sei();

	l *= (R5_VALUE + R6_VALUE) * 110;		// 分圧抵抗比と基準電圧 1.1V
	l /= R6_VALUE;
	l /= 1024;
	
	return (int)l;
}


//*******************************************************
//	SPIデータ出力
//*******************************************************
void SPI_out(uchar data){

	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}


//*******************************************************
//	時間待ち
//*******************************************************
void wait_timer(uchar t){

	Timer = t;
	while(Timer){
		WDR;
		SLEEP;
	}
}
